// decodeCCF.cpp
//

#include "stdafx.h"
#include "decodeCCF.hpp"

HINSTANCE hDll = NULL;
typeDecodeIR decodeIR;

FILE *inputFile;
FILE *outputFile;
unsigned char *ccf;
uint ccfLength;

int format = 0;
uint capability;
uint attribPos;

char empty[] = "";

char* sTop = empty;
char* sDevice = empty;
char* sPanel = empty;
char* sKey = empty;
char* sSequence = empty;
char* sFunction = empty;
char* sUdb = empty;
char* sDecode = empty;
char bDecode[256];

char* sProtocol = empty;
char bProtocol[256];
char* sDev = empty;
char bDev[256];
char* sSub = empty;
char bSub[256];
char* sObc = empty;
char bObc[256];
char* sEfc = empty;
char bEfc[256];
char* sMisc = empty;
char bMisc[256];

char* sFreq = empty;
char bFreq[256];
char* sOnce = empty;
char bOnce[256];
char* sRept = empty;
char bRept[256];

static int bursts[] =
{
    0,-0xCA0,      0x20,-0x20,   -0x20,0x20,                                                                                         // 00
    0x60,-0x20,    0x20,-0x20,   -0x20,0x20,   -0x10,0x10,   0x10,-0x10,   0,-0xBC0,                                                 // 01
    0,-0xAA0,      0,-0x80,      0x20,-0x20,   -0x20,0x20,                                                                           // 02
    0x5B,-0xC422,  0x5B,-0x1BC0, 0x5B,-0x1622, 0x5B,-0x1083, 0x5B,-0xAE4,  0x5B,-0x546,                                              // 03
    0x103,-0xAD01, 0x103,-0x47D5,0x103,-0x1745,0x1010,-0x403,0x103,-0x309, 0x103,-0x103, 0x1010,-0x81E,                              // 04
    0x227B,-0x1160,0x22C,-0x8FFA,0x22C,-0x6108,0x22C,-0x4165,0x22C,-0x31AF,0x22C,-0x2C13,0x22C,-0x684, 0x22C,-0x22C,0x227B,-0x1160,  // 05
    0xFC,-0x347B,  0xFC,-0xFC,   0xFC,-0x7E,   0x7E,-0x347B, 0x7E,-0x274,  0x7E,-0x17A,  0x7E,-0xFC,   0x7E,-0x7E,                   // 06,07

    0,0x19,-0x5A,0x28, -0x144,0x24,0,0, -0x36,0x12,-0xE,0x12, -0x22,0x12,-0x22,0x12, -0xE,0x12,-0x36,0x12, -0x7FC,0,0,0             // 08,09

};

static unsigned char burstN[] = { 3, 6, 4, 6, 7, 9, 0, 8, 0, 6 };
int* burstP[0x1E];

void initBursts()
{
	int *bp = bursts;
    for (int ndx=0; ndx<sizeof(burstN); ndx++)
    {
        burstP[ndx] = bp;
        bp += burstN[ndx];
        if (ndx==9)
            bp += burstN[ndx];
    }
}

static const char* hardKeys[][3] =
{
	{ "{action}","{action}","{action}" },
	{ "52",		"60",		"84" },
	{ "56",		"64",		"88" },
	{ "{Vol-}",	"{Vol-}",	"{Mute}" },
	{ "{Vol+}",	"{Vol+}",	"{Ch-}" },
	{ "{Ch-}",	"{Ch-}",	"{Ch+}" },
	{ "{Ch+}",	"{Ch+}",	"{Vol-}" },
	{ "{Mute}",	"{Mute}",	"{Vol+}" },
	{ "060",	"68",		"{M}" },
	{ 0,		"72",		"{Return}" },
	{ 0,		"076",		"{Ex}" },
	{ 0,		0,			"{Up}" },
	{ 0,		0,			"{Left}" },
	{ 0,		0,			"{Right}" },
	{ 0,		0,			"{Down}" },
	{ 0,		0,			"092" }
};

struct filterClass
{
    const char* tbl[256];
    char data[4096];
    char result[4096];
    const char* operator[](unsigned char nn)
    {
        return tbl[nn];
    }
    const char* operator()(char* p)
    {
        for (char* r=result; *p; ++p)
        {
            const char* t = (*this)[*p];
            int l = strlen(t);
            memcpy(r, t, l);
            r += l;
        }
        *r = 0;
        return result;
    }
    filterClass()
    {
        char *tmp = data;
        for (int nn=0; nn<256; ++nn)
        {
            tbl[nn] = tmp;
            if (nn>=' ' && nn<0x7E && nn!='`')
            {
                *tmp++ = nn;
            }
            else
            {
                sprintf(tmp, "`%02X", nn);
                tmp += 3;
            }
            *tmp++ = 0;
        }
	    tbl[0xA5] = "`Power_Toggle" ;
	    tbl[0x86] = "`Stop";
	    tbl[0x85] = "`Pause" ;
	    tbl[0x87] = "`Play" ;
	    tbl[0x89] = "`Reverse_Play";
	    tbl[0xAB] = "`Eject" ;
	    tbl[0xA2] = "`Step_Forward" ;
	    tbl[0xBE] = "`Step_Back" ;
	    tbl[0xBA] = "`Track_Forwards" ;
	    tbl[0xAA] = "`Track_Back" ;
	    tbl[0xA0] = "`Scan_Back" ;
	    tbl[0xB3] = "`Scan_Forward" ;
	    tbl[0xA7] = "`Down_Arrow";
	    tbl[0xAC] = "`Up_Arrow";
	    tbl[0x95] = "`Left_Arrow" ;
	    tbl[0x97] = "`Right_Arrow" ;
    }
} filter;

static const char* devTypes[] = 
{
    "A. Proc",
    "Cable",
    "CD",
    "CDR",
    "DBS",
    "DVD",
    "FlatTV",
    "Home",
    "LD",
    "PreAmp",
    "Tape",
    "Tuner",
    "TV",
    "VCR",
    "WebTV"
};

int main(int argc, char* argv[])
{
    initBursts();
	if (argc!=2)
	{
		fprintf(stderr,"use: decodeCCF filename\n");
		return 1;
	}

	const char* inputName = argv[1];
	const char* dot = strrchr(inputName, '.');
	if (dot == 0)
		dot = inputName + strlen(inputName);

	int len = dot-inputName;
	char *outputName = new char[len+5];
	memcpy( outputName, inputName, len);
	memcpy( outputName+len, ".txt", 5);

	int result = decodeCCF(inputName, outputName);
	return result;
}

int decodeCCF(const char* inputName, const char*outputName)
{
	if ( ! loadDll() )
	{
		printf("Failed to load DecodeIR.dll\n");
		exit(-1);
	}
	inputFile = fopen(inputName, "rb");
	if ( ! inputFile )
	{
		printf("Can't open input file: \"%s\"\n", inputName);
		exit(-1);
	}

	ccfLength = filelength(fileno(inputFile));

	ccf = new unsigned char[ccfLength + 1024];
	uint bytes = fread(ccf, 1, ccfLength, inputFile);
	fclose(inputFile);
	if (bytes != ccfLength)
	{
		printf("Error reading %d bytes from %s\nResult = %d\n", ccfLength, inputName, bytes);
		exit(-1);
	}

	outputFile = fopen(outputName, "w");
	if ( ! outputFile )
	{
		printf("Can't create output file: \"%s\"\n", outputName);
		exit(-1);
	}

	decodeCcf();

	fclose(outputFile);
	return 0;
}

void decodeCcf()
{
	// fprintf(outputFile, "Version = \"%s\"\n", getString(0));

	attribPos = getLong( 44 );
	capability = getLong( 36 );

	if ( isNewMarantz() )
		format = 2;
	else if ( hasColor() )
		format = 1;

	/* fprintf(outputFile, "Capability = %04X (", capability);
	if ( hasColor() )
		fprintf(outputFile, " color");
	if ( hasTimers() )
		fprintf(outputFile, " timers");
	if ( capability & 4 )
		fprintf(outputFile, " udb");
	if ( isNewMarantz() )
		fprintf(outputFile, " marantz-x200");
	fprintf(outputFile, " )\n");
	*/

	// fprintf(outputFile, " attribPos=0x%X\n", attribPos);

	fprintf(outputFile, "Protocol\tDevice\tSubDev\tOBC\tEFC\tMisc\tDecode\tUDB\tFunction\tSequence\tKey\tPanel\tDevName\tTop");
    fprintf(outputFile, "Freq\tOnce\tRept\n");
	sTop = "H";
	device( 48 );
	sTop = empty;
	device( 52 );
	sTop = "M";
	device( 56 );
    // panel( attribPos+4 );
}

void device(uint addr)
{
	for (;;)
	{
		uint ptr = getLong(addr);
		if (ptr == 0)
			break;
		
		sDevice = getString(ptr+4);

		for (int ndx = 0; ; ++ndx)
		{
			const char *name = hardKeys[ndx][format];
			if ( name[0] == '0' )
			{
				int pn = atoi(name);
				panel(ptr+pn);
				break;
			}
			if ( name[0] > '0' && name[0] <= '9' )
				name = getString( ptr + atoi(name) );

			actionList( ptr+16+4*ndx, name );
		}

		addr = ptr;
	}
	sDevice = empty;
}

void panel(uint addr)
{
	for (;;)
	{
		uint ptr = getLong(addr);
		if (ptr == 0)
			break;
		sPanel = getString(ptr + 4);
		int c1 = getByte(ptr+8);
		int c2 = getByte(ptr+9);
		if (c1 != c2)
			error("ERROR Panel Child Count mismatch %d,%d", c1, c2 );
		else
			for (int ndx=0; ndx<c1; ++ndx)
				child(ptr+10+9*ndx);
		addr = ptr;
	}
	sPanel = empty;
}

void child(uint addr)
{
	int ty=getByte(addr+8);
	switch (ty)
	{
	case 0:
		frame(addr+4);
		break;
	case 1:
		button(addr+4);
		break;
	default:
		error("ERROR panelChildType %d", ty);
	}
}

void frame(uint addr)
{
	uint ptr = getLong(addr);
	if (ptr)
	{
		ptr += hasColor() ? 21 : 18;
		int c1 = getByte(ptr);
		int c2 = getByte(ptr+1);
		if (c1 != c2)
			error("ERROR Frame Child Count mismatch %d,%d", c1, c2 );
		else
			for (int ndx=0; ndx<c1; ++ndx)
				child(ptr+2+9*ndx);
	}
}

void button(uint addr)
{
	uint ptr = getLong(addr);
	if (ptr)
	{
		char* key = getString(ptr+8);
		actionList(ptr+4, key);
	}
}

void actionList(uint addr, const char* name)
{
	uint ptr = getLong(addr);
	if (ptr == 0)
		return;
	sKey = const_cast<char*>(name);
	int count = getByte(ptr);
	if ( count != getByte(ptr+1) )
	{
		error("ERROR actionList Count mismatch %d,%d", count, getByte(ptr+1) );
	}
	for (int ndx=0; ndx<count; ndx++)
	{
		if (count > 1)
		{
			static char number[16];
			sprintf(number, "%d", ndx+1);
			sSequence = number;
		}
		action( ptr+2 + 9*ndx );
	}
	sSequence = empty;
	sKey = empty;
}

void action(uint addr)
{
	if ( getByte(addr) == 1 )
	{
		uint p1 = getLong( addr+1 );
		uint p2 = getLong( addr+5 );
		if (p1)
			error("ERROR action_1(%08X,%08X)", p1, p2);
		else
			ir(p2);
	}
}

void ir(uint addr)
{
	uint len = getWord( addr );
	sFunction = getString( addr+2 );
	uint ptr = addr+6;
	if ( hasUdb() )
	{
		static char bufr[64];
		int u1 = getWord(ptr);
		int u2 = getWord(ptr+2);
		int u3 = getWord(ptr+4);
		if (u1 || u2 || u3)
			sprintf(bufr,"%X.%X.%X", u1, u2, u3 );
		ptr += 6;
		sUdb = bufr;
	}
	uint kind = getWord( ptr );
	uint wave = getWord( ptr+2 );
	uint len1 = getWord( ptr+4 );
	uint len2 = getWord( ptr+6 );
	ptr += 8;
	if ( len != (len1+len2)*4 + ptr-addr || ((len1 | len2)&-1024))
		error ("ERROR IR length %d(%X,%d,%d)", len, kind, len1, len2);
	else
	{
        len = len1+len2;
		switch (kind)
		{
		case 0:
		case 1:
			callDll(kind, wave, len1, len2, ptr);
			break;
		case 0x5000:
			rc5(len, ptr);
			break;
		case 0x5001:
			rc5x(len, ptr);
			break;
		case 0x6000:
			rc6(len, ptr);
			break;
		case 0x7000:
			d7000(len, ptr);
			break;
		case 0x8000:
			udb(len, ptr);
			break;
        case 0x9000:
            fourBytes("NEC1", len, ptr);
            break;
        case 0x900A:
            twoWords("NEC1", len, ptr);
            break;
        case 0x900B:
            twoWords("NEC2", len, ptr);
            break;
        case 0x900E:
            pioneer2(len, ptr);
            break;
		}
		if ( sProtocol == empty )
		{
            len *= 4;
            sprintf( sProtocol=bProtocol, "%04X %04X %04X %04X", kind, wave, len1, len2);
            for (uint ndx=0; ndx<len && ndx<12; ndx+=2)
                sprintf( sProtocol+strlen(sProtocol), " %04X", getWord(ptr+ndx) );
            show();
		}
		sProtocol = empty;
	}
	sUdb = empty;
	sFunction = empty;
}

void udb(uint len, uint addr)
{
	if (len >= 2)
	{
        int devType = getWord(addr);
        if ( devType > 0 && devType <= 15 )
    		sprintf( sProtocol=bProtocol, "udb-%s", devTypes[devType-1] );
        else
    		sprintf( sProtocol=bProtocol, "udb-%d", devType );
		sprintf( sDev=bDev, "%d", getWord(addr+2) );
		sprintf( sObc=bObc, "%d", getWord(addr+4) );
		if (len > 2)
			sprintf( sMisc=bMisc, "{len=%d}", len);
		show();
	}
}

void rc5(uint len, uint addr)
{
	if (len)
	{
		sProtocol="RC5";
		sprintf( sDev=bDev, "%d", getWord(addr) );
		sprintf( sObc=bObc, "%d", getWord(addr+2) );
		if (len > 1)
			sprintf( sMisc=bMisc, "{len=%d}", len);
		show();
	}
}

void rc5x(uint len, uint addr)
{
	if (len>1)
	{
		sProtocol="RC5x";
		sprintf( sDev=bDev, "%d", getWord(addr) );
		sprintf( sSub=bSub, "%d", getWord(addr+2) );
		sprintf( sObc=bObc, "%d", getWord(addr+4) );
		if (len != 2)
			sprintf( sMisc=bMisc, "{len=%d}", len);
		show();
	}
}

void rc6(uint len, uint addr)
{
	if (len)
	{
		sProtocol="RC6";
		sprintf( sDev=bDev, "%d", getWord(addr) );
		sprintf( sObc=bObc, "%d", getWord(addr+2) );
		if (len > 1)
			sprintf( sMisc=bMisc, "{len=%d}", len);
		show();
	}
}

void d7000(uint len, uint addr)
{
	if (len > 2)
	{
        uint dId = getWord(addr);
        uint count = getWord(addr+2);
        if ( (count+3)>>1 == len )
        {
            uint d0 = getWord(addr+4);
            if ( dId == 0xC && count == 0xE && d0==0x10 && d7000Thomson(addr+6) )
                return;
            if ( dId == 0xB && count == 0xD && d0==0x10 && d7000RECS80b(addr+6) )
                return;
                
        }
    }
    sprintf( sProtocol=bProtocol, "7000.%X.%X", getWord(addr-6), len );
    if ( len > 0 )
    {
        uint count =  getWord(addr+2);
        sprintf( sDev=bDev, "%X.%X", getWord(addr), count );
        if ( count > (len-1)<<1 )
            count = (len-1)<<1;
        char* ss = sSub = bSub;
        int f = -1;
        for (uint ndx = 0; ndx<count; ++ndx)
        {
            uint v = getWord(addr+4+2*ndx);
            if (f>=0 && v>15)
                *(ss++) = '/';
            else if (f>0)
                *(ss++) = '\\';
            sprintf(ss, "%X", v);
            ss += strlen(ss);
            f = ( v > 15 );
        }
    }
    show();
}

int d7000Thomson(uint addr)
{
    int dev=0;
    int obc=0;
    for (int ndx=0; ndx<12; ++ndx)
    {
        uint bt = getWord(addr+2*ndx)-3;
        if (ndx==4)
        {
            if (bt != 0x19)
                return 0;
        }
        else
        {
            if (bt > 1)
                return 0;
            if (bt == 0)
            {
                if (ndx < 6)
                    dev += ( 1 << ndx );
                else
                    obc += ( 1 << (ndx-6) );
            }
        }
    }
    sProtocol = "Thomson";
    sprintf(sDev=bDev, "%d", dev);
    sprintf(sObc=bObc, "%d", obc);
    sprintf(sEfc=bEfc, "%03d_%03d", hex2efc( lsb(63-obc) ), hex2efc( lsb(127-(obc*2+(dev>>5))) ) );
    sprintf(sMisc=bMisc,"{7000 C.E} L=%d", getWord(addr+24) );
    show();
    return 1;
}

int d7000RECS80b(uint addr)
{
    if (getWord(addr) != 0x16 || getWord(addr+2) != 0x1A)
        return 0;
    int dev=0;
    int obc=0;
    for (int ndx=0; ndx<9; ++ndx)
    {
        uint bt = getWord(addr+4+2*ndx)-2;
        if (bt > 1)
            return 0;
        if (bt == 0)
        {
            if (ndx < 3)
                dev += ( 8 >> ndx );
            else
                obc += ( 0x100 >> ndx );
        }
    }
    sProtocol = "RECS80";
    sprintf(sDev=bDev, "%d", dev);
    sprintf(sObc=bObc, "%d", obc);
    sprintf(sEfc=bEfc, "%03d", hex2efc( (63-obc)<<2 ) );
    sprintf(sMisc=bMisc,"(479/5773/8675) {7000 B.D} L=%d", getWord(addr+22) );
    sFreq = "33.4286";
    show();
    sFreq = empty;
    return 1;
}

int d7000RECS80a(uint addr)
{
    if (getWord(addr) != 0x2 || getWord(addr+2) != 0x17)
        return 0;
    int dev=0;
    int obc=0;
    for (int ndx=0; ndx<9; ++ndx)
    {
        uint bt = getWord(addr+4+2*ndx)-2;
        if (bt > 1)
            return 0;
        if (bt == 0)
        {
            if (ndx < 3)
                dev += ( 8 >> ndx );
            else
                obc += ( 0x100 >> ndx );
        }
    }
    sProtocol = "RECS80";
    sprintf(sDev=bDev, "%d", dev);
    sprintf(sObc=bObc, "%d", obc);
    sprintf(sEfc=bEfc, "%03d", hex2efc( (63-obc)<<2 ) );
    sprintf(sMisc=bMisc,"{7000 A.D} L=%d", getWord(addr+22) );
    sFreq = "33.4286";
    show();
    sFreq = empty;
    return 1;
}

void fourBytes(char* protocol, uint len, uint addr)
{
    if (len >= 2)
    {
        int obc = getWord(addr+4);
        if ( obc + getWord(addr+6) == 255 )
        {
            sProtocol = protocol;
            sprintf( sMisc=bMisc, "{%04X.%X}", getWord(addr-8), getWord(addr-6) );
            if ( len > 2 )
                sprintf( sMisc+strlen(sMisc), " {len=%d}", len );
            int dev = getWord(addr);
		    sprintf( sDev=bDev, "%d", dev );
            int sub = getWord(addr+2);
            if ( dev + sub != 255 )
    		    sprintf( sSub=bSub, "%d", sub );
		    sprintf( sObc=bObc, "%d", obc );
			sprintf( sEfc=bEfc, "%03d", hex2efc( lsb(255-obc) ) );
    		show();
        }

    }
}

void pioneer2(uint len, uint addr)
{
    if (len >= 2)
    {
        int wave = getWord(addr-6);
        int obc[2];
        obc[0] = getByte(addr+2);
        obc[1] = getByte(addr+6);
        if ( 255-getByte(addr+3) == obc[0] && 255-getByte(addr+7) == obc[1] && wave>=0x66 && wave<=0x6A )
        {
            sprintf( bMisc, "{%04X.%X}", getWord(addr-8), getWord(addr-6) );
            if ( len > 2 )
                sprintf( bMisc+strlen(bMisc), " {len=%d}", len );
            for (int ndx=0; ndx<2; ++ndx)
            {
                sProtocol = "Pioneer";
                sMisc = bMisc;
		        sprintf( sDecode=bDecode, "%d_34", ndx*34 );
                int dev = getByte(addr+ndx*4);
		        sprintf( sDev=bDev, "%d", dev );
                int sub = getByte(addr+1+ndx*4);
                if ( dev + sub != 255 )
    		        sprintf( sSub=bSub, "%d", sub );
		        sprintf( sObc=bObc, "%d", obc[ndx] );
			    sprintf( sEfc=bEfc, "%03d", hex2efc( lsb(255-obc[ndx]) ) );
    		    show();
            }
        }

    }
}

void twoWords(char* protocol, uint len, uint addr)
{
    if (len >= 1)
    {
        int obc = getByte(addr+2);
        if ( obc + getByte(addr+3) == 255 )
        {
            sProtocol = protocol;
            sprintf( sMisc=bMisc, "{%04X.%X}", getWord(addr-8), getWord(addr-6) );
            if ( len > 1 )
                sprintf( sMisc+strlen(sMisc), " {len=%d}", len );
            int dev = getByte(addr);
		    sprintf( sDev=bDev, "%d", dev );
            int sub = getByte(addr+1);
            if ( dev + sub != 255 )
    		    sprintf( sSub=bSub, "%d", sub );
		    sprintf( sObc=bObc, "%d", obc );
			sprintf( sEfc=bEfc, "%03d", hex2efc( lsb(255-obc) ) );
    		show();
        }

    }
}

int getByte(uint addr)
{
	return (addr<ccfLength) ? ccf[addr] : 0;
}

int getWord(uint addr)
{
	return (getByte(addr)<<8) + getByte(addr+1);
}

int getLong(uint addr)
{
	return (getWord(addr)<<16) + getWord(addr+2);
}

char* getString(uint addr)
{
	uint addr2 = getLong(addr) & 0x7FFFFFFF;
	uint len = getByte(addr2);
	return getString(addr2+1, len);
}

char* getString(uint addr, uint len)
{
	if (addr==1)
		return empty;
	if (len+addr < ccfLength)
	{
		char* result = new char[len+1];
		memcpy(result, ccf+addr, len);
		result[len]=0;
		return result;
	}
	char* result = new char[50];
	sprintf(result, "[string ERROR %04X:%d]", addr, len);
	return result;
}

void show()
{
	fprintf(outputFile, "%s", filter(sProtocol) );
	fprintf(outputFile, "\t%s", filter(sDev) );
	fprintf(outputFile, "\t%s", filter(sSub) );
	fprintf(outputFile, "\t%s", filter(sObc) );
	fprintf(outputFile, "\t%s", filter(sEfc) );
	fprintf(outputFile, "\t%s", filter(sMisc) );
	fprintf(outputFile, "\t%s", filter(sDecode) );

	sDev = sSub = sObc = sEfc = sMisc = sDecode = empty;

	fprintf(outputFile, "\t%s", filter(sUdb) );
	fprintf(outputFile, "\t%s", filter(sFunction) );
	fprintf(outputFile, "\t%s", filter(sSequence) );
	fprintf(outputFile, "\t%s", filter(sKey) );
	fprintf(outputFile, "\t%s", filter(sPanel) );
	fprintf(outputFile, "\t%s", filter(sDevice) );
	fprintf(outputFile, "\t%s", filter(sTop) );

	fprintf(outputFile, "\t%s", filter(sFreq) );
	fprintf(outputFile, "\t%s", filter(sOnce) );
	fprintf(outputFile, "\t%s\n", filter(sRept) );
}

void error(char* fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vsprintf(bProtocol, fmt, ap );
	sProtocol = bProtocol;
	show();
	sProtocol = empty;
}

int loadDll() 
{
	hDll = LoadLibrary("DecodeIR.dll");
	if (hDll != NULL)
	{
		decodeIR = (typeDecodeIR)GetProcAddress(hDll, "DecodeIR");
		if (decodeIR)
		{
			return 1;
		}
		FreeLibrary(hDll);
	}
	return 0;
}

int callDll(rawData& in, decodeData& d)
{
	d.protocol[0] = 0;
	d.misc[0] = 0;
	d.error[0] = 0;
	d.device = -1;
	d.subDevice = -1;
	d.obc = -1;
	static int hex[]={-1,-1,-1,-1};
	memcpy(d.hex, hex, sizeof(hex));

	decodeIR(d.context, in.raw, (int)(in.freq*1000.), in.once, in.again,
		d.protocol, &d.device, &d.subDevice, &d.obc, d.hex, d.misc, d.error);

	return d.protocol[0];
}

void callDll(int unmodulated, int wave, int len1, int len2, uint addr)
{
	int len =(len1 + len2)*2;
	if (wave==0 || len < 8)
		return;
	rawData raw;
	decodeData decodes[20];

	double unit = wave * .241246;

	raw.freq = unmodulated ? 0 : 1000./unit;
	raw.once = len1;
	raw.again = len2;

    sprintf(sFreq=bFreq, "%g", raw.freq);
    sprintf(sOnce=bOnce, "%d", raw.once);
    sprintf(sRept=bRept, "%d", raw.again);

	raw.raw = new int[len];

	while ( --len >= 0)
		raw.raw[len] = (int)(getWord(addr+2*len)*unit);

	memset(decodes[0].context, 0, 2*sizeof(int) );
	int count;
	for (count=0; ; )
	{
		if ( ! callDll( raw, decodes[count] ) )
			break;
        if ( ++count == 20 )
            break;
		memcpy(decodes[count].context, decodes[count-1].context, 2*sizeof(int));
	}

	for (int ndx=0; ndx<count; ++ndx)
	{
		sProtocol = decodes[ndx].protocol;
		if ( decodes[ndx].device >= 0 )
			sprintf( sDev=bDev, "%d", decodes[ndx].device );
		if ( decodes[ndx].subDevice >= 0 )
			sprintf( sSub=bSub, "%d", decodes[ndx].subDevice );
		if ( decodes[ndx].obc >= 0 )
			sprintf( sObc=bObc, "%d", decodes[ndx].obc );
		if ( decodes[ndx].hex[0] >= 0 )
		{
			sprintf( sEfc=bEfc, "%03d", hex2efc( decodes[ndx].hex[0] ) );
			for (int hn=1; decodes[ndx].hex[hn] >= 0; ++hn)
				sprintf( bEfc+strlen(bEfc), "_%03d", hex2efc( decodes[ndx].hex[hn] ) );
		}
		if ( decodes[ndx].misc[0] || decodes[ndx].error[0] )
		{
			sprintf( sMisc=bMisc, "%s", decodes[ndx].misc );
			if ( decodes[ndx].error[0] )
				sprintf( bMisc+strlen(bMisc), " {%s}", decodes[ndx].error );
		}
		if ( count > 1 )
		{
			sprintf( sDecode=bDecode, "%d_%d", decodes[ndx].context[0], (decodes[ndx].context[1]>>16)+5 );
		}
		show();
	}

    sFreq = sOnce = sRept = empty;
	delete raw.raw;
}