#pragma once

typedef unsigned int uint;

int decodeCCF(const char* inputName, const char*outputName);
void decodeCcf();
void device(uint addr);
void panel(uint addr);
void child(uint addr);
void frame(uint addr);
void button(uint addr);
void actionList(uint addr, const char* name);
void action(uint addr);
void ir(uint addr);
void udb(uint len, uint addr);
void rc5(uint len, uint addr);
void rc5x(uint len, uint addr);
void rc6(uint len, uint addr);
void d7000(uint len, uint addr);
int d7000Thomson(uint addr);
int d7000RECS80a(uint addr);
int d7000RECS80b(uint addr);
void fourBytes(char* protocol, uint len, uint addr);
void twoWords(char* protocol, uint len, uint addr);
void pioneer2(uint len, uint addr);
void callDll(int unmodulated, int wave, int len1, int len2, uint addr);

int getByte(uint addr);
int getWord(uint addr);
int getLong(uint addr);
char* getString(uint addr);
char* getString(uint addr, uint len);
extern uint capability;

inline int hasColor() {return capability & (1<<8); }
inline int hasTimers() {return capability & (1<<17); }
inline int hasUdb() {return capability & (1<<18); }
inline int isNewMarantz() {return capability & (1<<11); }
inline int isOldMarantz() {return capability == 1; }
inline int isMarantz() {return isOldMarantz() || isNewMarantz(); }

void show();
void error(char* fmt, ...);


typedef void (CALLBACK* typeDecodeIR) (unsigned int* Context,
		int* bursts, 	
		const int freq, 
		const int singleBurstCount, 
		const int repeatBurstCount, 
		char* protocol, 
		int* device, 
		int* subDevice, 
		int* OBC, 
		int hex[4], 
		char* misc, 
		char* error);
int loadDll();

struct rawData
{
	double freq;
	int once;
	int again;
	int* raw;
};

struct decodeData
{
	uint context[2];
	int device;
	int subDevice;
	int obc;
	int hex[4];
	char protocol[128];
	char misc[128];
	char error[128];
};

inline int hex2efc(int hex)
{
	// Rotate left three places  (ignoring affect on bits 8-10)
	hex = (hex<<3) + (hex>>5);

	// XOR with 0x51;
	hex ^= 0xAE;

	// Add 100 (ignoring affect on bits 8-10)
	hex += 100;

	// Return only low 8 bits
	return hex & 0xFF;
}

inline uint lsb(uint n)
{
    uint t = (n>>4) + (n<<4);
    t      = ((t>>2)&0x33) + ((t&0x33)<<2);
    return   ((t>>1)&0x55) + ((t&0x55)<<1);
}